using ITensors

# TensorNetworkCircuit struct and public functions
export TensorNetworkCircuit, add_input!, add_output!, qubits
export contract_tn!, decompose_tensor!, create_test_tnc
export input_tensors, output_tensors, input_indices, output_indices, single_amplitude

const input_output_tensors = Dict{Char,Array{Float64, 1}}('0' => [1., 0.],
                                                          '1' => [0., 1.],
                                                          '+' => [1., 1.]./sqrt(2),
                                                          '-' => [1., -1.]./sqrt(2))

"""Tensor network circuit data-structure"""
mutable struct TensorNetworkCircuit
    qubits::Int64
    input_indices::Array{Index, 1}
    output_indices::Array{Index, 1}
    input_tensors::Array{Union{Symbol, Nothing}, 1}
    output_tensors::Array{Union{Symbol, Nothing}, 1}
    tn::TensorNetwork
end

function TensorNetworkCircuit(qubits::Int64)
    input_indices = [Index(2, tags="Qubit $i") for i in 1:qubits]
    output_indices = copy(input_indices)
    input_tensors = Array{Union{Symbol, Nothing}, 1}(nothing, qubits)
    output_tensors = Array{Union{Symbol, Nothing}, 1}(nothing, qubits)
    TensorNetworkCircuit(qubits, input_indices, output_indices, input_tensors, output_tensors, TensorNetwork())
end

function Base.copy(tnc::TensorNetworkCircuit)
    tn = TensorNetwork()
    f = x -> x === nothing ? nothing : push!(tn, tnc[x]; tid=x)
    my_input_tensors = convert(Array{Union{Symbol, Nothing}}, f.(input_tensors(tnc)))
    my_output_tensors = convert(Array{Union{Symbol, Nothing}}, f.(output_tensors(tnc)))
    for x in keys(tnc)
        if !(x in input_tensors(tnc)) && !(x in output_tensors(tnc))
            push!(tn, tnc[x]; tid=x)
        end
    end
    tn.next_id = tnc.tn.next_id

    TensorNetworkCircuit(qubits(tnc),
                         copy(input_indices(tnc)),
                         copy(output_indices(tnc)),
                         my_input_tensors,
                         my_output_tensors,
                         tn)
end

Base.eltype(::TensorNetworkCircuit) = ITensor
Base.getindex(tnc::TensorNetworkCircuit, i::Symbol) = tnc.tn[i]
Base.getindex(tnc::TensorNetworkCircuit, i::T) where T <: Index = tnc.tn[i]
Base.iterate(tnc::TensorNetworkCircuit) = iterate(values(tnc))
Base.iterate(tnc::TensorNetworkCircuit, i) = iterate(values(tnc), i)
Base.keys(tnc::TensorNetworkCircuit) = keys(tnc.tn)
Base.length(tnc::TensorNetworkCircuit) = length(tnc.tn)
Base.show(io::IO, ::MIME"text/plain", tnc::TensorNetworkCircuit) = print(io, "TensorNetworkCircuit(qubits => $(qubits(tnc)), gates => $(length(tnc)))")
Base.values(tnc::TensorNetworkCircuit) = values(tnc.tn)

input_tensors(tnc::TensorNetworkCircuit) = tnc.input_tensors
output_tensors(tnc::TensorNetworkCircuit) = tnc.output_tensors
input_indices(tnc::TensorNetworkCircuit) = tnc.input_indices
output_indices(tnc::TensorNetworkCircuit) = tnc.output_indices
bonds(tnc::TensorNetworkCircuit) = bonds(tnc.tn)
qubits(tnc::TensorNetworkCircuit) = tnc.qubits
tensor_data(tnc::TensorNetworkCircuit, i; kwargs...) = tensor_data(tnc.tn, i; kwargs...)
hyperindices(tnc::TensorNetworkCircuit, i; kwargs...) = hyperindices(tnc.tn, i; kwargs...)
get_hyperedges(tnc::TensorNetworkCircuit) = get_hyperedges(tnc.tn)
disable_hyperindices!(tnc::TensorNetworkCircuit) = disable_hyperindices!(tnc.tn)


"""
    push!(tnc::TensorNetworkCircuit,
          qubits::Vector{Int64},
          data::Array{T, 2}) where T

Function to add a gate to the tensor network circuit given
the qubits it acts on and an array of the matrix elements
"""
function Base.push!(tnc::TensorNetworkCircuit,
                    qubits::Vector{Int64},
                    data::Array{T, 2};
                    decompose::Bool=true) where T
    gate_input_indices = tnc.output_indices[qubits]
    tnc.output_indices[qubits] = gate_output_indices = [prime(x) for x in gate_input_indices]
    if length(qubits) == 1 || !decompose
        indices = [gate_output_indices..., gate_input_indices...]
        @assert prod(size(data)) == prod(dim.(indices)) "Data matrix dimension does not match indices"
        data = reshape(data, Tuple(dim.(indices)))
        push!(tnc, indices, data)
    elseif length(qubits) == 2 && decompose
        data = reshape(data, Tuple([dim.(gate_output_indices)..., dim.(gate_input_indices)...]))
        A, B = decompose_gate(data)
        if size(A)[3] == size(B)[1] == 1
            A = reshape(A, Tuple(size(A)[1:2]))
            B = reshape(A, Tuple(size(B)[2:3]))
            a_indices = [gate_output_indices[1], gate_input_indices[1]]
            b_indices = [gate_output_indices[2], gate_input_indices[2]]
        else
            virtual_index = Index(size(A)[3])
            a_indices = [gate_output_indices[1], gate_input_indices[1], virtual_index]
            b_indices = [virtual_index, gate_output_indices[2], gate_input_indices[2]]
        end
        push!(tnc, a_indices, A)
        push!(tnc, b_indices, B)
    else

        @error("Gates spanning more than two qubits not supported with decomposition yet.")
    end
end

Base.push!(tnc::TensorNetworkCircuit, indices::Vector{<:Index}, data::AbstractArray; kwargs...) = push!(tnc.tn, indices, data; kwargs...)

"""
    delete!(tnc::TensorNetworkCircuit, tensor_id::Symbol)

Function to remove a tensor from a tensor network circuit.
"""
function Base.delete!(tnc::TensorNetworkCircuit, tensor_id::Symbol)
    replace!(tnc.input_tensors, tensor_id => nothing)
    replace!(tnc.output_tensors, tensor_id => nothing)
    delete!(tnc.tn, tensor_id)
end

"""
    push_input!(tnc::TensorNetworkCircuit, tensor::Array{Elt, 1}, pos::Int64) where Elt

Function to add a single input tensor to the tensor network circuit at the given position
"""
function push_input!(tnc::TensorNetworkCircuit, tensor::Array{Elt, 1}, pos::Int64) where Elt
    index = tnc.input_indices[pos]
    tensor_sym = tnc.input_tensors[pos]
    if tensor_sym !== nothing
        delete!(tnc, tensor_sym)
    end
    tnc.input_tensors[pos] = push!(tnc, [index], tensor; tid=tensor_sym)
end

"""
    push_output!(tnc::TensorNetworkCircuit, tensor::Array{Elt, 1}, pos::Int64) where Elt

Function to add a single output tensor to the tensor network circuit at the given position
"""
function push_output!(tnc::TensorNetworkCircuit, tensor::Array{Elt, 1}, pos::Int64) where Elt
    index = tnc.output_indices[pos]
    tensor_sym = tnc.output_tensors[pos]
    if tensor_sym !== nothing
        delete!(tnc, tensor_sym)
    end
    tnc.output_tensors[pos] = push!(tnc, [index], tensor; tid=tensor_sym)
end

"""
    add_input!(tnc::TensorNetworkCircuit; input::Union{String, Nothing}=nothing)

Function to add input tensors to the circuit
"""
function add_input!(tnc::TensorNetworkCircuit, input::Union{String, Nothing}=nothing)
    if input === nothing input = "0"^qubits(tnc) end
    [push_input!(tnc, input_output_tensors[input[pos]], pos) for pos in 1:qubits(tnc)]
end

"""
    add_output!(tnc::TensorNetworkCircuit; output::Union{String, Nothing}=nothing)

Function to add output tensors to the circuit
"""
function add_output!(tnc::TensorNetworkCircuit, output::Union{String, Nothing}=nothing)
    if output === nothing output = "0"^qubits(tnc) end
    [push_output!(tnc, input_output_tensors[output[pos]], pos) for pos in 1:qubits(tnc)]
end

contract_tn!(tnc::TensorNetworkCircuit, plan) = contract_tn!(tnc.tn, plan)

contract_tn(tnc::TensorNetworkCircuit, plan) = contract_tn!(copy(tnc), plan)

function decompose_tensor!(tnc::TensorNetworkCircuit, args...; kwargs...)
    decompose_tensor!(tnc.tn, args...; kwargs...)
end

simple_contraction(tnc::TensorNetworkCircuit) = simple_contraction(tnc.tn)
simple_contraction!(tnc::TensorNetworkCircuit) = simple_contraction!(tnc.tn)

"""
    create_test_tnc(;input::Union{String, Nothing}=nothing,
                    output::Union{String, Nothing}=nothing,
                    no_input::Bool=false,
                    no_output::Bool=false,
                    kwargs...)

Create a tensor network circuit for a small example circuit, 3 qubit ghz preparation circuit in this case
"""
function create_test_tnc(;input::Union{String, Nothing}=nothing,
                         output::Union{String, Nothing}=nothing,
                         no_input::Bool=false,
                         no_output::Bool=false,
                         kwargs...)
    tnc = TensorNetworkCircuit(3)
    if !no_input add_input!(tnc, input) end
    push!(tnc, [1], Gates.h(); kwargs...)
    push!(tnc, [2, 1], Gates.cx(); kwargs...)
    push!(tnc, [3, 2], Gates.cx(); kwargs...)
    if !no_output add_output!(tnc, output) end
    tnc
end

